#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>
#include <iostream>

const std::string SUBPACKAGE      = "voms-proxy-info";
//const std::string VERSION         = "0.1";
const std::string CONFILENAME     = "";
const std::string USERCONFILENAME = "";

extern "C" {
#include <globus_gsi_credential.h>
#include <globus_gsi_system_config.h>
/*
#ifdef USE_PKCS11
#include "scutils.h"
#endif
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

}

#include "options.h"

#include "../api/voms_api.h"

bool debug = false;
bool quiet = false;

#define MAXVOS 2000
static std::vector<char *>voms;
static std::vector<char *>order;
static std::string vo;


static std::string        certfile;
static std::string        keyfile;
static std::string        proxyfile;
static std::string        certdir;
static std::string        certcafile;
static std::string        pin         = "";
static std::string        argp;
static std::string        program;
static std::string        conffile    = CONFILENAME;
static std::string        userconf    = "";
static std::string        incfile     = "";
static int                version     = 0;
static bool               verify      = false;
static bool               progversion = false;
static bool               ignorewarn  = false;
static bool               failonwarn  = false;
static int                vomslife    = -1;
static std::string        separate    = "";
#ifdef CLASS_ADD
static void *		  class_add_buf = NULL;
static size_t		  class_add_buf_len = 0;
#endif
extern "C" {
static int (*pw_cb)() = NULL;
};


int
main(int argc, char **argv)
{
    struct rlimit newlimit = {0,0};
    
    if (strrchr(argv[0],'/'))
	program = strrchr(argv[0],'/') + 1;
    else
	program = argv[0];
  
    static char *LONG_USAGE = \
      "\n" \
      "    Options\n" \
      "    -help, -usage           Displays usage\n" \
      "    -version                Displays version\n" \
      "    -debug                  Enables extra debug output\n" \
      "    -q                      Quiet mode, minimal output\n" \
      "    -verify                 Verifies certificate to make proxy for\n" \
      "    -cert     <certfile>    Non-standard location of user certificate\n" \
      "    -key      <keyfile>     Non-standard location of user key\n" \
      "    -certdir  <certdir>     Non-standard location of trusted cert dir\n" \
      "    -file     <proxyfile>   Non-standard location of proxy cert\n" \
      "    -ignorewarn             Ignore warnings.\n"
      "    -failonwarn             Treat warnings as errors.\n"
      "\n";

    set_usage(LONG_USAGE);

    struct option opts[] = {
      {"help",       0, NULL,                OPT_HELP},
      {"usage",      0, NULL,                OPT_HELP},
      {"version",    0, (int *)&progversion, OPT_BOOL},
      {"cert",       1, (int *)&certfile,    OPT_STRING},
      {"certdir",    1, (int *)&certdir,     OPT_STRING},
      {"file",       1, (int *)&proxyfile,   OPT_STRING},
      {"key",        1, (int *)&keyfile,     OPT_STRING},
      {"debug",      0, (int *)&debug,       OPT_BOOL},
      {"verify",     0, (int *)&verify,      OPT_BOOL},
      {"q",          0, (int *)&quiet,       OPT_BOOL},
      {"quiet",      0, (int *)&quiet,       OPT_BOOL},
      {"ignorewarn", 0, (int *)&ignorewarn,  OPT_BOOL},
      {"failonwarn", 0, (int *)&failonwarn,  OPT_BOOL},
      {0, 0, 0, 0}
    };

    if (!getopts(argc, argv, opts))
      exit(1);

    if (failonwarn && ignorewarn)
      ignorewarn = false;

    if (progversion) {
      std::cout << SUBPACKAGE << "\nVersion: " << VERSION << std::endl;
      std::cout << "Compiled: " << __DATE__ << " " << __TIME__ << std::endl;
      exit(0);
    }

//    if (setrlimit(RLIMIT_CORE, &newlimit) != 0)
//      exit(1);

  vomsdata voms;
  X509 *cert = NULL;
  STACK_OF(X509) *chain = NULL;
  const char* proxy_filename = NULL; 
  globus_gsi_cred_handle_t handle = NULL;

  globus_module_activate(GLOBUS_GSI_CREDENTIAL_MODULE);
  globus_module_activate(GLOBUS_GSI_SYSCONFIG_MODULE);

  if(!proxyfile.empty()) proxy_filename=proxyfile.c_str();

  if(!proxy_filename) {
    if(GLOBUS_GSI_SYSCONFIG_GET_PROXY_FILENAME((char**)&proxy_filename,
                  GLOBUS_PROXY_FILE_INPUT) != GLOBUS_SUCCESS) {
      std::cerr<<"Failed to find proxy file"<<std::endl;
      goto err;
    };
  };

  if(globus_gsi_cred_handle_init(&handle,GLOBUS_NULL) != GLOBUS_SUCCESS) {
    std::cerr<<"Failed to initialize credentials"<<std::endl;
    goto err;
  };
  
  if(globus_gsi_cred_read_proxy(handle,(char*)proxy_filename) != GLOBUS_SUCCESS) {
    std::cerr<<"Failed to read proxy"<<std::endl;
    goto err;
  };

  if(globus_gsi_cred_get_cert(handle,&cert) != GLOBUS_SUCCESS) {
    std::cerr<<"Failed to retrieve proxy's certificate"<<std::endl;
    goto err;
  };

  if(globus_gsi_cred_get_cert_chain(handle,&chain) != GLOBUS_SUCCESS) {
    std::cerr<<"Failed to retrieve certificates' chain"<<std::endl;
    goto err;
  };

  if(!voms.Retrieve(cert,chain)) {
    std::cerr<<"Failed to retrieve VOMS information from proxy"<<std::endl;
    goto err;
  };

  if(!voms.workvo.empty()) std::cerr<<"VO: "<<voms.workvo<<std::endl;
  if(!voms.extra_data.empty()) std::cerr<<"Extra: "<<voms.extra_data<<std::endl;
  for(int n = 0;n<voms.data.size();n++) {
    std::cerr<<"VOMS data ("<<n<<")"<<std::endl;
    if(!voms.data[n].user.empty())
       std::cerr<<"  User: "<<voms.data[n].user<<std::endl;
    if(!voms.data[n].userca.empty())
       std::cerr<<"  User's CA: "<<voms.data[n].userca<<std::endl;
    if(!voms.data[n].server.empty())
       std::cerr<<"  Server: "<<voms.data[n].server<<std::endl;
    if(!voms.data[n].serverca.empty())
       std::cerr<<"  Server's CA: "<<voms.data[n].serverca<<std::endl;
    if(!voms.data[n].voname.empty())
       std::cerr<<"  VO: "<<voms.data[n].voname<<std::endl;
    if(!voms.data[n].uri.empty())
       std::cerr<<"  URI: "<<voms.data[n].uri<<std::endl;
    const char* s;
    if(!voms.data[n].date1.empty()) {
       s=voms.data[n].date1.c_str();
       s=strrchr(s,'\r'); if(s) std::cerr<<"  Valid from: "<<s+1<<std::endl;
    };
    if(!voms.data[n].date2.empty()) {
       s=voms.data[n].date2.c_str();
       s=strrchr(s,'\r'); if(s) std::cerr<<"  Valid till: "<<s+1<<std::endl;
    };
    if(!voms.data[n].custom.empty())
       std::cerr<<"  Extra: "<<voms.data[n].custom<<std::endl;
    for(int nn = 0;nn<voms.data[n].std.size();nn++) {
      std::cerr<<"  Characteristics ("<<nn<<"):"<<std::endl;
      if(!voms.data[n].std[nn].group.empty())
         std::cerr<<"    Group: "<<voms.data[n].std[nn].group<<std::endl;
      if(!voms.data[n].std[nn].role.empty())
         std::cerr<<"    Role: "<<voms.data[n].std[nn].role<<std::endl;
      if(!voms.data[n].std[nn].cap.empty())
         std::cerr<<"    Capabilities: "<<voms.data[n].std[nn].cap<<std::endl;
    };
  };


  globus_module_deactivate(GLOBUS_GSI_SYSCONFIG_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
  exit(0);

err:
  std::cerr<<"Failed."<<std::endl;
  globus_module_deactivate(GLOBUS_GSI_SYSCONFIG_MODULE);
  globus_module_deactivate(GLOBUS_GSI_CREDENTIAL_MODULE);
  exit(1);

}


